package ru.amse.baltijsky.javascheme.exporter.image;

import ru.amse.baltijsky.javascheme.tree.SchemaTree;
import ru.amse.baltijsky.javascheme.tree.walker.DepthFirstTreeWalker;
import ru.amse.baltijsky.javascheme.tree.walker.ITreeWalker;
import ru.amse.baltijsky.javascheme.tree.walker.TraversalState;
import ru.amse.baltijsky.javascheme.tree.walker.TraversalState.Rel;
import ru.amse.baltijsky.javascheme.util.Dir;
import ru.amse.baltijsky.javascheme.util.Size;
import static ru.amse.baltijsky.javascheme.util.Util.max;
import ru.amse.baltijsky.javascheme.util.VisConst;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;

/**
 * Exports a schema tree to image by drawing it in the specified graphic context.
 * Before a tree is drawn, its visual representation is created using <tt>VisualTreeBuilder</tt>.
 */
class ImageExporter implements IImageExporter {
    /**
     * Direction shorthands.
     */
    private static final int L = Dir.l();
    private static final int R = Dir.r();
    private static final int U = Dir.u();
    private static final int D = Dir.d();
    /**
     * Root of the visual tree.
     */
    private VisualNode visualTree = null;
    /**
     * The schema tree to export to image.
     */
    private SchemaTree schemaTree;
    /**
     * Determines whether to draw the frames around node subtrees.
     */
    private boolean drawSubtreeFrames = true;

    /**
     * Creates a new image exporter that visualizes the given schema tree.
     *
     * @param theSchemaTree the schema tree to export to image
     */
    public ImageExporter(SchemaTree theSchemaTree) {
        if (theSchemaTree == null) {
            throw new IllegalArgumentException("Schema tree must be non-null");
        }
        schemaTree = theSchemaTree;
    }

    /**
     * Draws the schema tree image on the specified graphics with the left upper corner (= left upper corner of the
     * bounding rectangle), shifting at the specified position in th grpahics.
     * Graphics is also used to obtain current font metrics, which is necessary to calculate the sizes.
     *
     * @param g      graphics where to draw
     * @param whereX x-coordinate of the left corners
     * @param whereY y-coordinate of the upper corners
     */
    public void draw(Graphics g, int whereX, int whereY) {
        if (visualTree == null) {
            visualTree = (new VisualTreeBuilder()).build(schemaTree, g);
        }

        // adjust drawing position so that its left corner will be (whereX, whereY)
        whereX += visualTree.getLeftSubtreeSize();

        ITreeWalker<VisualNode> walker = new DepthFirstTreeWalker<VisualNode>(visualTree);
        int lastDepth = 0;
        // joint points stack for the subtree-blocks of the nodes that have children
        Deque<Size> jointPts = new LinkedList<Size>();
        jointPts.push(new Size(whereX, whereY));
        // stack of queues for children chains of the same if's (for if-else constructions) leading line x-coord
        Deque<Queue<Integer>> childrenHeadsX = new LinkedList<Queue<Integer>>();
        // stack of y-coords of the first real nodes (not just branch shapes) of the children of the current node
        Deque<Integer> childrenHeadsY = new LinkedList<Integer>();
        // current drawing point
        Size cur = new Size(whereX, whereY);
        // last drawn node
        VisualNode lastNode = null;

        Color oldColor = g.getColor();
        g.setColor(VisConst.SCHEME_COL);
        while (walker.hasNextState()) {
            TraversalState<VisualNode> state = walker.getNextState();

            drawClosingLevels(lastNode, cur, state.depth, lastDepth, jointPts, g);
            if (state.rel == Rel.NEXT || state.rel == Rel.ROOT) {
                if (state.rel == Rel.NEXT) {
                    // draw transition from the previous node
                    g.drawLine(cur.x, cur.y, cur.x, cur.y + state.node.getOffset().y);
                }
                drawNodeAndChildrenHeads(state.node, cur, g, jointPts, childrenHeadsX, childrenHeadsY);
            } else if (state.rel == Rel.CHILD) {
                cur.x = childrenHeadsX.peek().remove();
                cur.y = childrenHeadsY.peek();
                if (childrenHeadsX.peek().isEmpty()) {
                    childrenHeadsX.pop();
                    childrenHeadsY.pop();
                }
            }

            lastDepth = state.depth;
            lastNode = state.node;
        }

        // close levels for the rest of the nodes
        drawClosingLevels(lastNode, cur, 0, lastDepth, jointPts, g);
        g.setColor(oldColor);
    }


    /**
     * Draws transitions for currently closing levels.
     *
     * @param lastNode  the last drawn node
     * @param cur       current coordinates
     * @param depth     current depth in the visual tree traversal
     * @param lastDepth last drawn node depth in the visual tree
     * @param jointPts  stack of joint points for nodes that have children
     * @param g         the graphics context where everything is drawn
     */
    private void drawClosingLevels(VisualNode lastNode, Size cur, int depth, int lastDepth, Deque<Size> jointPts, Graphics g) {
        VisualNode p = lastNode;
        for (int i = lastDepth - depth; p != null && i > 0; --i) {
            p = p.getParent();
            int[] closingShapeDims = p.getClosingShape().getShapeDims(g);
            // connect the already-drawn children tails; the x-coordinate for the rightmost (last) tail is cur.x
            Size jointPt = jointPts.pop();
            cur.y = jointPt.y + p.getChildrenSubtreeSize().y;
            g.drawLine(jointPt.x + closingShapeDims[R], cur.y, cur.x, cur.y);
            // draw closing shape and transition to the next node
            cur.x = jointPt.x;
            p.getClosingShape().draw(cur.x, cur.y, g);
            cur.y += closingShapeDims[D];
            g.drawLine(cur.x, cur.y, cur.x, cur.y + VisConst.YGAP);
            cur.y += VisConst.YGAP;
        }
    }

    /**
     * Draws the specified node and heads for its children.
     *
     * @param node           the node to draw
     * @param cur            current drawing point
     * @param g              graphics context
     * @param jointPts       joint points stack for node's parents
     * @param childrenHeadsX children heads stack to update with the node's children heads
     * @param childrenHeadsY
     */
    private void drawNodeAndChildrenHeads(VisualNode node, Size cur, Graphics g, Deque<Size> jointPts, Deque<Queue<Integer>> childrenHeadsX, Deque<Integer> childrenHeadsY) {
        int[] shapeDim = node.getShape().getShapeDims(g);
        int ySubtreeBegin = cur.y;

        drawSubtreeFrame(node, cur, g);
        drawNode(node, shapeDim, cur, g);
        drawTransitions(node, shapeDim, jointPts, cur, ySubtreeBegin, g);
        drawChildrenHeads(node, jointPts, cur, ySubtreeBegin, childrenHeadsX, childrenHeadsY, g);
    }

    private void drawSubtreeFrame(VisualNode node, Size cur, Graphics g) {
        if (drawSubtreeFrames && node.hasChildren()) {
            Color oldColor = g.getColor();
            //        g.drawRect(cur.x - node.getLeftSubtreeSize(), cur.y, node.getChildrenSubtreeSize().x, node.getChildrenSubtreeSize().y);
            g.setColor(VisConst.SCHEME_FRAME_COL);
            //g.drawRect(cur.x - node.getLeftSubtreeSize(), cur.y, node.getSubtreeSize().x, node.getChildrenSubtreeSize().y + node.getClosingShape().getShapeDims(g)[D] + VisConst.YGAP);
            int leftOffset = max(node.getShape().getShapeDims(g)[L], node.getClosingShape().getShapeDims(g)[L]) + VisConst.XGAP;
            int bottomMargin = node.getClosingShape().getShapeDims(g)[U] + VisConst.YGAP;
            g.drawRect(cur.x - leftOffset, cur.y, node.getChildrenSubtreeSize().x, node.getChildrenSubtreeSize().y - bottomMargin);
            g.setColor(oldColor);
        }
        if (node == visualTree) {
            Color oldColor = g.getColor();
            //        g.drawRect(cur.x - node.getLeftSubtreeSize(), cur.y, node.getChildrenSubtreeSize().x, node.getChildrenSubtreeSize().y);
            g.setColor(VisConst.SCHEME_FRAME_COL);
            g.drawRect(cur.x - node.getLeftSubtreeSize(), cur.y, node.getSubtreeSize().x, node.getSubtreeSize().y);
            g.setColor(oldColor);
        }
    }

    private void drawNode(VisualNode node, int[] shapeDim, Size cur, Graphics g) {
        cur.y += node.getOffset().y + shapeDim[U];
        node.getShape().draw(cur.x, cur.y, g);
    }

    private void drawTransitions(VisualNode node, int[] shapeDim, Deque<Size> jointPts, Size cur, int ySubtreeBegin, Graphics g) {
        int[] closingShapeDims = node.getClosingShape().getShapeDims(g);
        if (node.hasNext()) {
            // draw transition to the next node
            g.drawLine(cur.x, cur.y + shapeDim[D], cur.x, ySubtreeBegin + node.getChildrenSubtreeSize().y - closingShapeDims[U]);
        } else if (node.hasParent()) {
            // draw tail to the level closure skipping the closing node
            VisualNode p = node.getParent();
            int subtreeStartY = cur.y - node.getShape().getShapeDims(g)[U] - node.getOffset().y;
            int closingShapeStartY = subtreeStartY + node.getChildrenSubtreeSize().y - closingShapeDims[U];
            int closingShapeEndY = subtreeStartY + node.getChildrenSubtreeSize().y + closingShapeDims[D];
            g.drawLine(cur.x, cur.y + shapeDim[D], cur.x, closingShapeStartY);
            g.drawLine(cur.x, closingShapeEndY, cur.x, jointPts.peek().y + p.getChildrenSubtreeSize().y);
        }
    }

    private void drawChildrenHeads(VisualNode node, Deque<Size> jointPts, Size cur, int ySubtreeBegin, Deque<Queue<Integer>> childrenHeadsX, Deque<Integer> childrenHeadsY, Graphics g) {
        if (node.hasChildren()) {
            jointPts.push(new Size(cur.x, ySubtreeBegin));
            Queue<Integer> newChildrenHeads = new LinkedList<Integer>();
            cur.x = drawChildren(node, cur, g, newChildrenHeads, childrenHeadsY);
            childrenHeadsX.push(newChildrenHeads);
        } else if (node.hasNext()) {
            cur.y = ySubtreeBegin + node.getChildrenSubtreeSize().y;
        } else {
            // proceed to the next level
            if (node.hasParent()) {
                cur.y = jointPts.peek().y + node.getParent().getChildrenSubtreeSize().y;
            }
        }
    }

    /**
     * Draws the children of the specified node and returns x-coordinate of the centre the first child of the node.
     * Also updates the queue for x-coordinates of other children's centres (heads).
     *
     * @param node              the node which children to draw
     * @param cur               coordinates of the node for which to draw the children
     * @param g                 graphics context
     * @param newChildrenHeadsX updated children heads x-coordinates
     * @param childrenHeadsY    y-coordinates of children heads
     * @return x-coordinate of the center of the first child node
     */
    private int drawChildren(VisualNode node, Size cur, Graphics g, Queue<Integer> newChildrenHeadsX, Deque<Integer> childrenHeadsY) {
        // add the first child in a special way
        Iterator<VisualNode> childIt = node.getChildren();
        VisualNode child = childIt.next();
        int[] firstChildDim = child.getShape().getShapeDims(g);
        int shapeDimR = node.getShape().getShapeDims(g)[R];
        int closeShapeDimR = node.getClosingShape().getShapeDims(g)[R];
        int firstChildX = cur.x + max(
                max(shapeDimR, closeShapeDimR) + 2 * VisConst.XGAP + firstChildDim[L],
                child.getLeftSubtreeSize());
        g.drawLine(cur.x + shapeDimR, cur.y, firstChildX - firstChildDim[L], cur.y);
        child.getShape().draw(firstChildX, cur.y, g);

        // start of the next child's subtree
        childrenHeadsY.push(cur.y + firstChildDim[D]);

        // add the rest of the children
        newChildrenHeadsX.add(firstChildX);
        drawRestChildren(child, childIt, newChildrenHeadsX, cur, firstChildX, g);

        return firstChildX;
    }

    private void drawRestChildren(VisualNode child, Iterator<VisualNode> childIt, Queue<Integer> newChildrenHeadsX, Size cur, int childX, Graphics g) {
        VisualNode prev = child;
        for (; childIt.hasNext();) {
            child = childIt.next();
            int newChildX = childX + prev.getRightSubtreeSize() + child.getLeftSubtreeSize();
            g.drawLine(
                    childX + prev.getShape().getShapeDims(g)[R], cur.y,
                    newChildX - child.getShape().getShapeDims(g)[L], cur.y);
            child.getShape().draw(newChildX, cur.y, g);
            childX = newChildX;
            newChildrenHeadsX.add(childX);
            prev = child;
        }
    }

    public Size getImageSize(Graphics g) {
        if (visualTree == null) {
            visualTree = (new VisualTreeBuilder()).build(schemaTree, g);
        }
        return visualTree.getSubtreeSize();
    }

    public String getDescription() {
        return schemaTree.getDescription();
    }

    public void exportToFile(File f, String formatName) throws IOException, IllegalArgumentException {
        // learn the size of the image to draw
        // image0 is used only to obtain Graphics that can render font
        BufferedImage image0 = new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB);
        Graphics sizeCalculator = image0.createGraphics();
        sizeCalculator.setFont(VisConst.FONT);
        int w = getImageSize(sizeCalculator).x + 1;
        int h = getImageSize(sizeCalculator).y + 1;

        // create the image of the schema tree
        BufferedImage image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        Graphics g = image.createGraphics();
        ((Graphics2D) (g)).addRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON));
        ((Graphics2D) (g)).addRenderingHints(new RenderingHints(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB));
        g.setFont(VisConst.FONT);
        g.setColor(Color.WHITE);
        g.fillRect(0, 0, w, h);

        draw(g, 0, 0);

        if (!ImageIO.write(image, formatName, f)) {
            throw new IOException("could not write the image in " + formatName + "  format");
        }
    }

    public void setFramesSubtreeDecoration(boolean value) {
        drawSubtreeFrames = value;
    }
}
