package ru.amse.baltijsky.javascheme.exporter.stream;

import ru.amse.baltijsky.javascheme.tree.BranchNode;
import ru.amse.baltijsky.javascheme.tree.SchemaNode;
import ru.amse.baltijsky.javascheme.tree.walker.ITreeWalker;
import ru.amse.baltijsky.javascheme.tree.walker.TraversalState;

import java.io.PrintWriter;
import java.util.*;

/**
 * A dumb exporter that takes the nodes returned during the loadTree tree traversal by <code>TreeWalker</code> and
 * outputs them to the specified output stream.
 * The user may call the <code>export</code> method several times (in that sense the objects are reentrant).
 */
public class TextExporter implements IStreamExporter {
    /**
     * A non-reentrant implementation of the exporter.
     * Created each time when <code>export</code> is called.
     */
    private class TextExporterImpl {
        /**
         * Symbols used to designate the current schema tree level.
         */
        private final static String INDENT = " |";
        /**
         * Printer to the output stream.
         */
        private PrintWriter out;
        /**
         * A walker that is used to obtain the nodes of the loadTree tree in the appropriate order.
         */
        private ITreeWalker<SchemaNode> treeWalker;
        /**
         * Current traversal state.
         */
        private TraversalState<SchemaNode> state;
        /**
         * Conditions of the branches for all currently existing schema tree levels.
         */
        private Deque<Iterator<List<String>>> branchesConditions = new LinkedList<Iterator<List<String>>>();
        /**
         * The level of the last node that was printed out.
         */
        private int lastLevel = 0;
        /**
         * Builder for the leading blank string; is changed every time the schema tree traversal level changes.
         */
        private StringBuilder leadBlank = new StringBuilder(INDENT);

        /**
         * The constructor.
         *
         * @param newOut        the output stream
         * @param newTreeWalker the walker
         */
        public TextExporterImpl(PrintWriter newOut, ITreeWalker<SchemaNode> newTreeWalker) {
            out = newOut;
            treeWalker = newTreeWalker;
            treeWalker.reset();
        }

        /**
         * Exports the tree to the output stream.
         */
        void export() {
            while (treeWalker.hasNextState()) {
                state = treeWalker.getNextState();
                for (int i = state.depth - lastLevel; i > 0; --i) {
                    // add indentation
                    leadBlank.append(INDENT);
                }
                for (int i = state.depth - lastLevel; i < 0; ++i) {
                    // remove indentation
                    leadBlank.delete(leadBlank.length() - INDENT.length(), leadBlank.length());
                }
                exportSingleLine();
                lastLevel = state.depth;
            }
        }

        /**
         * Prints a node to the standard output with all the code in single line.
         */
        private void exportSingleLine() {
            writeCondition();
            printBracketed("--", state.node.getNodeType().getDescription());

            // print out the code if it's present
            out.println(getCode());

            printChildlessIfCondition();
        }

        /**
         * Gets the code for the current node in the schema tree traversal.
         *
         * @return code for the current node
         */
        private String getCode() {
            String[] code = null;
            if (state.node.getCode() != null && (state.node.getCode().length != 0) && !state.node.getCode()[0].equals("")) {
                code = state.node.getCode();
            }
            StringBuilder sb = new StringBuilder();
            if (code != null) {
                sb.append("[");
                for (String s : code) {
                    sb.append(" ").append(s);
                }
                sb.append("]");
            }

            return sb.toString();
        }

        /**
         * Handles condition printing for if/else branch if current node is of that kind and setting new
         * conditions iterator if a new If is encountered.
         */
        private void writeCondition() {
            if ((state.rel == TraversalState.Rel.CHILD) && (state.prev instanceof BranchNode)) {
                out.print(leadBlank + "==" + Arrays.toString(branchesConditions.peek().next().toArray()));
                out.println();
                if (!branchesConditions.peek().hasNext()) {
                    branchesConditions.pop();
                }
            }
            if (state.node instanceof BranchNode) {
                BranchNode branchNode = (BranchNode) state.node;
                if (branchNode.hasChildren()) {
                    branchesConditions.push(((BranchNode) state.node).getConditions());
                }
            }
        }

        /**
         * Prints condition of a childless if-statement.
         * It is done separately because for a normal if-statement conditions are printed out before its children.
         */
        private void printChildlessIfCondition() {
            if (state.node instanceof BranchNode && !state.node.hasChildren() && ((BranchNode) state.node).hasConditions()) {
                out.print(leadBlank + INDENT + "==" + Arrays.toString(((BranchNode) state.node).getConditions().next().toArray()));
                out.println();
            }
        }

        /**
         * Prints the specified string, indented, enclosed in square brackets with the perfix before the brackets.
         *
         * @param prefix the prefix
         * @param s      the string to print out
         */
        private void printBracketed(String prefix, String s) {
            out.print(leadBlank + prefix + "[" + s + "]");
        }
    }

    /**
     * Exports to the standard output.
     *
     * @see IStreamExporter
     */
    public void export(PrintWriter out, ITreeWalker<SchemaNode> treeWalker) {
        (new TextExporterImpl(out, treeWalker)).export();
    }
}