package ru.amse.baltijsky.javascheme.exporter.stream;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import ru.amse.baltijsky.javascheme.tree.NodeType;
import ru.amse.baltijsky.javascheme.tree.SchemaNode;
import ru.amse.baltijsky.javascheme.tree.walker.ITreeWalker;
import ru.amse.baltijsky.javascheme.tree.walker.TraversalState;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;


/**
 * Exports the internal data to an XML file in XMI format.
 */
public class XmiExporter implements IStreamExporter {
    /**
     * The XML document that is created and then exported.
     */
    private Document doc;

    /**
     * A helper class to simplify assigning the most common values of attributes to the elements of the XML DOM tree.
     */
    private class Attr {
        /**
         * Attr constant.
         */
        public final static int XMI_ID = 1;
        /**
         * Attr constant.
         */
        public final static int IS_SPEC = 1 << 1;
        /**
         * Attr constant.
         */
        public final static int IS_ROOT = 1 << 2;
        /**
         * Attr constant.
         */
        public final static int IS_LEAF = 1 << 3;
        /**
         * Attr constant.
         */
        public final static int IS_ABSTRACT = 1 << 4;
        /**
         * Attr constant.
         */
        public final static int IS_ACTIVE = 1 << 5;

        /**
         * Sets the attributes specified via a mask for the specified <code>XML DOM Element</code>.
         *
         * @param e     the element to set the attributes
         * @param attrs attributes mask: the attributes to set
         */
        private void setCommonAttributesValue(Element e, int attrs) {
            if ((attrs & Attr.IS_ACTIVE) != 0) {
                e.setAttribute("isActive", "false");
            }
            if ((attrs & Attr.IS_ABSTRACT) != 0) {
                e.setAttribute("isAbstract", "false");
            }
            if ((attrs & Attr.IS_LEAF) != 0) {
                e.setAttribute("isLeaf", "false");
            }
            if ((attrs & Attr.IS_ROOT) != 0) {
                e.setAttribute("isRoot", "false");
            }
            if ((attrs & Attr.IS_SPEC) != 0) {
                e.setAttribute("isSpecification", "false");
            }
            if ((attrs & Attr.XMI_ID) != 0) {
                e.setAttribute("xmi.id", generateNextXmiId());
            }
        }
    }

    /**
     * The <code>Attr</code> instance.
     */
    public final Attr attribute = new Attr();

    public XmiExporter() throws ParserConfigurationException {
        doc = createDocument();
    }

    public void export(PrintWriter out, ITreeWalker<SchemaNode> treeWalker) {
        try {
            Element root = createRootNode();

            // create XMI header
            createXmiHeader(root);
            // create contents of the diagram
            Element model = createDiagramHeader(root, "model1");

            Element ownedElement = doc.createElement("UML:Namespace.ownedElement");
            model.appendChild(ownedElement);
            Element umlClass = createUmlCLass(ownedElement);

            // add the traits of the method that we're going to output (return type, arguments) as children of the ownedElement\
            // add the parameter types
            addMethodTraits(ownedElement, "void");
            // add the return type
            Element returnType = addMethodTraits(ownedElement, "int");

            // create a branch of ownedElement that is responsible for the specialities of the class
            Element classifierFeature = doc.createElement("UML:Classifier.feature");
            umlClass.appendChild(classifierFeature);
            // the function for which the activity diagram is created
            Element func = createOperation(classifierFeature);
            // specify the function parameter type (referencing the previously created node holding the type)
            createFunctionParamType(func, returnType);

            // create another branch of umlClass containing the diagram itself
            ownedElement = doc.createElement("UML:Namespace.ownedElement");
            umlClass.insertBefore(ownedElement, classifierFeature);
            // set the context of the diagram (reference to the function it belongs to)
            Element activityGraph = createActivityGraph(ownedElement, func);
            // write the diagram itself
            createDiagram(activityGraph, treeWalker);

            // now output the XML to the PrintStream via Transformer
            outputXml(out);
        } catch (TransformerConfigurationException e) {
            e.printStackTrace();
        } catch (TransformerException e) {
            e.printStackTrace();
        }
    }

    private void outputXml(PrintWriter out) throws TransformerException {
        TransformerFactory trFactory = TransformerFactory.newInstance();
        Transformer tr = trFactory.newTransformer();
        tr.setOutputProperty(OutputKeys.INDENT, "yes");
        tr.setOutputProperty(OutputKeys.STANDALONE, "no");
        tr.transform(new DOMSource(doc), new StreamResult(out));
    }

    private void createDiagram(Element activityGraph, ITreeWalker<SchemaNode> treeWalker) {
        // write the states
        Element stateMachineTop = doc.createElement("UML:StateMachine.top");
        activityGraph.appendChild(stateMachineTop);
        Element compositeState = doc.createElement("UML:CompositeState");
        compositeState.setAttribute("name", "top");
        compositeState.setAttribute("isConcurrent", "false");
        attribute.setCommonAttributesValue(compositeState, Attr.XMI_ID | Attr.IS_SPEC);
        stateMachineTop.appendChild(compositeState);

        // we'll export the diagram in two ways  writing vertices of different kinds and transitions between them.
        // so we'll add children to subVertex ...
        Element subVertex = doc.createElement("UML:CompositeState.subvertex");
        compositeState.appendChild(subVertex);
        // ... and to transitions
        Element transitions = doc.createElement("UML:StateMachine.transitions");
        activityGraph.appendChild(transitions);

        // prepare the tree walker
        treeWalker.reset();

        // skip nodes that are hierarchically superior to the first node in the function (translation unit, class, etc)
        // and remember the root of the function activity diagram
        SchemaNode rootNode = null;
        while (treeWalker.hasNextState()) {
            rootNode = treeWalker.getNextState().node;
            NodeType t = rootNode.getNodeType();
            if (t == NodeType.METHOD) {
                break;
            }
        }
        if (rootNode != null) {
            // current element
            Element initialElement = doc.createElement("UML:Pseudostate");
            initialElement.setAttribute("kind", "initial");
            attribute.setCommonAttributesValue(initialElement, Attr.XMI_ID | Attr.IS_SPEC);
            subVertex.appendChild(initialElement);
            // map of previously traversed nodes to the created elements to add the transitions
            Map<SchemaNode, Element> prevElements = new HashMap<SchemaNode, Element>();
            Element curElement = null;
            while (treeWalker.hasNextState()) {
                // current traversal node
                TraversalState<SchemaNode> nextState = treeWalker.getNextState();
                SchemaNode curNode = nextState.node;
                SchemaNode prevNode = nextState.prev;
                curElement = createElementByType(curNode);
                Element prevElement = prevElements.get(prevNode);
                if (prevElement == null) {
                    prevElement = initialElement;
                }
                // add the node to the map
                prevElements.put(curNode, curElement);
                // add vertex
                subVertex.appendChild(curElement);

                // add transition
                addTransition(prevElement, curElement, transitions);
            }
            // add transition from the last action to the final state
            Element finalElement = doc.createElement("UML:FinalState");
            attribute.setCommonAttributesValue(finalElement, Attr.XMI_ID | Attr.IS_SPEC);
            subVertex.appendChild(finalElement);
            if (curElement != null) {
                addTransition(curElement, finalElement, transitions);
            }
        }
    }

    private void createDiagramContext(Element activityGraph, Element func) {
        Element context = doc.createElement("UML:StateMachine.context");
        activityGraph.appendChild(context);
        Element operationRef = doc.createElement("UML:Operation");
        operationRef.setAttribute("xmi.idref", func.getAttribute("xmi.id"));
        context.appendChild(operationRef);
    }

    private Element createActivityGraph(Element ownedElement, Element func) {
        // create activity graph header
        Element activityGraph = doc.createElement("UML:ActivityGraph");
        attribute.setCommonAttributesValue(activityGraph, Attr.XMI_ID | Attr.IS_SPEC);
        ownedElement.appendChild(activityGraph);
        // create diagram context
        createDiagramContext(activityGraph, func);
        return activityGraph;
    }

    private void createFunctionParamType(Element func, Element returnType) {
        Element bihevioralParameter = doc.createElement("UML:BehavioralFeature.parameter");
        func.appendChild(bihevioralParameter);
        Element parameter = doc.createElement("UML:Parameter");
        parameter.setAttribute("kind", "return");
        parameter.setAttribute("name", "return");
        attribute.setCommonAttributesValue(parameter, Attr.XMI_ID | Attr.IS_SPEC);
        bihevioralParameter.appendChild(parameter);
        Element parameter_type = doc.createElement("UML:Parameter.type");
        parameter.appendChild(parameter_type);
        Element dataTypeRef = doc.createElement("UML:DataType");
        dataTypeRef.setAttribute("xmi.idref", returnType.getAttribute("xmi.id"));
        parameter_type.appendChild(dataTypeRef);
    }

    private Element createOperation(Element classifierFeature) {
        Element func = doc.createElement("UML:Operation");
        func.setAttribute("concurrency", "sequential");
        func.setAttribute("isQuery", "false");
        func.setAttribute("ownerScope", "instance");
        func.setAttribute("visibility", "public");
        func.setAttribute("name", "func");
        attribute.setCommonAttributesValue(func, Attr.XMI_ID | Attr.IS_SPEC | Attr.IS_ROOT | Attr.IS_LEAF | Attr.IS_ABSTRACT);
        classifierFeature.appendChild(func);
        return func;
    }

    private Element createUmlCLass(Element ownedElement) {
        Element umlClass = doc.createElement("UML:Class");
        attribute.setCommonAttributesValue(umlClass, Attr.XMI_ID | Attr.IS_SPEC | Attr.IS_ROOT | Attr.IS_LEAF | Attr.IS_ABSTRACT | Attr.IS_ACTIVE);
        umlClass.setAttribute("visibility", "public");
        umlClass.setAttribute("name", "Dummy");
        ownedElement.appendChild(umlClass);
        return umlClass;
    }

    private Element addMethodTraits(Element ownedElement, String name) {
        Element trait = doc.createElement("UML:DataType");
        trait.setAttribute("name", name);
        attribute.setCommonAttributesValue(trait, Attr.XMI_ID | Attr.IS_SPEC | Attr.IS_LEAF | Attr.IS_ROOT | Attr.IS_ABSTRACT);
        ownedElement.appendChild(trait);
        return trait;
    }

    /**
     * Adds a transition between the two specified elements of the diagram.
     *
     * @param srcElement  the source of the transition
     * @param destElement the destination of the transition
     */
    private void addTransition(Element srcElement, Element destElement, Element transitions) {
        // create the transition
        Element newTransition = doc.createElement("UML:Transition");
        attribute.setCommonAttributesValue(newTransition, Attr.XMI_ID | Attr.IS_SPEC);

        // add source element to the transition
        Element src = doc.createElement("UML:Transition.source");
        addTransitionPoint(src, srcElement, newTransition);
        // add destination element
        Element dest = doc.createElement("UML:Transition.target");
        addTransitionPoint(dest, destElement, newTransition);

        transitions.appendChild(newTransition);

        // add references to the created transition to the current and previous elements
        // add outgoing reference to the previous node
        Element outgoing = doc.createElement("UML:StateVertex.outgoing");
        addTransitionRef(srcElement, outgoing, newTransition);
        // add incoming reference to the current node
        Element incoming = doc.createElement("UML:StateVertex.incoming");
        addTransitionRef(destElement, incoming, newTransition);
    }

    /**
     * Adds a transition reference (incoming or outgoing, the method isn't aware) to the specified diagram element.
     *
     * @param e          the diagram element where to put the transition reference
     * @param refKind    an element specifying whether the reference is incoming or outgoing (created by caller)
     * @param transition the transition reference to which is added
     */
    private void addTransitionRef(Element e, Element refKind, Element transition) {
        Element transitionRef = doc.createElement("UML:Transition");
        transitionRef.setAttribute("xmi.idref", transition.getAttribute("xmi.id"));
        refKind.appendChild(transitionRef);
        e.appendChild(refKind);
    }

    /**
     * Adds a new point to the transition (source or destination).
     *
     * @param point      the point to be specified and added (created by the caller)
     * @param elemToRef  element that is referenced in the point
     * @param transition the transition where to add the point
     */
    private void addTransitionPoint(Element point, Element elemToRef, Element transition) {
        Element elemRef = doc.createElement(elemToRef.getTagName());
        elemRef.setAttribute("xmi.idref", elemToRef.getAttribute("xmi.id"));
        point.appendChild(elemRef);
        transition.appendChild(point);
    }

    /**
     * Creates common header for the diagrams, including model
     */
    private Element createDiagramHeader(Element root, String modelName) {
        Element contents = doc.createElement("XMI.content");
        root.appendChild(contents);
        Element model = doc.createElement("UML:Model");
        model.setAttribute("name", modelName);
        attribute.setCommonAttributesValue(model, Attr.XMI_ID | Attr.IS_SPEC | Attr.IS_ROOT | Attr.IS_LEAF | Attr.IS_ABSTRACT);
        contents.appendChild(model);
        return model;
    }

    /**
     * Creates XMI header for the specified document (it is added as the child of root).
     *
     * @param root root element that will be the parent of the header
     */
    private void createXmiHeader(Element root) {
        Element header = doc.createElement("XMI.header");
        root.appendChild(header);
        Element metamodel = doc.createElement("XMI.metamodel");
        metamodel.setAttribute("xmi.version", "1.4");
        metamodel.setAttribute("xmi.name", "UML");
        header.appendChild(metamodel);
    }

    /**
     * Creates root element for the specified XMI document.
     *
     * @return the created root element
     */
    private Element createRootNode() {
        Element root = doc.createElement("XMI");
        root.setAttribute("xmi.version", "1.2");
        root.setAttribute("xmlns:UML", "org.omg.xmi.namespace.UML");
        doc.appendChild(root);
        return root;
    }

    /**
     * Creates a new XML DOM document to write our tree there.
     *
     * @return the created document
     * @throws ParserConfigurationException if the parser was not configured correctly
     */
    private static Document createDocument() throws ParserConfigurationException {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        DocumentBuilder builder = factory.newDocumentBuilder();
        return builder.newDocument();
    }

    private Element createElementByType(SchemaNode p) {
        Element ret;
        switch (p.getNodeType()) {
        default:
            ret = doc.createElement("UML:ActionState");
            attribute.setCommonAttributesValue(ret, Attr.XMI_ID | Attr.IS_SPEC);
            ret.setAttribute("isDynamic", "false");
            if (p.getCode().length > 0) {
                // write the code associated with the state
                Element entry = doc.createElement("UML:State.entry");

                Element uninterpretedAction = doc.createElement("UML:UninterpretedAction");
                uninterpretedAction.setAttribute("isAsynchronous", "false");
                attribute.setCommonAttributesValue(uninterpretedAction, Attr.XMI_ID | Attr.IS_SPEC);
                Element actionScript = doc.createElement("UML:Action.script");
                Element actionExpression = doc.createElement("UML:ActionExpression");
                attribute.setCommonAttributesValue(actionExpression, Attr.XMI_ID);
                actionExpression.setAttribute("body", p.getNodeType().getDescription()
                        + String.format("%n") + p.getCodeAsString());
                actionExpression.setAttribute("language", "");
                actionScript.appendChild(actionExpression);
                uninterpretedAction.appendChild(actionScript);
                entry.appendChild(uninterpretedAction);
                ret.appendChild(entry);
            }
        }
        return ret;
    }

    private String generateNextXmiId() {
        return Long.toString(System.nanoTime(), 16);
    }
}
