package ru.amse.baltijsky.javascheme.importer.test;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import ru.amse.baltijsky.javascheme.exporter.stream.TextExporter;
import ru.amse.baltijsky.javascheme.importer.IImporter;
import static ru.amse.baltijsky.javascheme.importer.test.TestImporterTools.*;
import ru.amse.baltijsky.javascheme.tree.*;
import static ru.amse.baltijsky.javascheme.tree.NodeType.*;
import ru.amse.baltijsky.javascheme.tree.walker.DepthFirstTreeWalker;
import ru.amse.baltijsky.javascheme.util.Util;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import java.io.*;
import java.util.ArrayList;
import java.util.List;

/**
 * Reads XML file with test desription and creates a schema tree to test visual representation module.
 */
public class TestImporter implements IImporter {
    private static final String SCHEMA_FILE = "test/testprops.xsd";

    public SchemaTree loadTree(Reader in, String className, String methodName) throws IOException {
        SchemaTree[] trees = loadTrees(in);
        return trees.length > 0 ? trees[0] : null;
    }

    public SchemaTree[] loadTrees(Reader in) throws IOException {
        try {
            Element testRoot = createDoc(in).getDocumentElement();
            NodeList children = testRoot.getChildNodes();
            int childrenAmount = children.getLength();

            List<SchemaTree[]> tests = new ArrayList<SchemaTree[]>();
            FirstLevelTestCreator firstLevelTestCreator = new FirstLevelTestCreator();
            int treesAmount = 0;
            for (int i = 0; i < childrenAmount; ++i) {
                Node xmlNode = children.item(i);
                if (xmlNode.getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }
                SchemaTree[] curTests = firstLevelTestCreator.createTests((Element) xmlNode);
                tests.add(curTests);
                treesAmount += curTests.length;
            }
            SchemaTree[] ret = new SchemaTree[treesAmount];
            int pos = 0;
            for (SchemaTree[] testBunch : tests) {
                System.arraycopy(testBunch, 0, ret, pos, testBunch.length);
                pos += testBunch.length;
            }
            return ret;
        } catch (SAXException e) {
            throw new IOException("Error in test properties file: " + e.getMessage());
        } catch (ParserConfigurationException e) {
            throw new IOException("Cannot parse test properties file: " + e.getMessage());
        } catch (IllegalArgumentException e) {
            throw new IOException("Error in test properties file: " + e.getMessage());
        }
    }

    public String[] getClasses(Reader in) throws IOException {
        return new String[0];
    }

    public String[] getMethods(Reader in, String className) throws IOException {
        return new String[0];
    }

    private Document createDoc(Reader in) throws SAXException, ParserConfigurationException, IOException {
        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        final SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        Schema schema = schemaFactory.newSchema(new File(SCHEMA_FILE));
        factory.setSchema(schema);
        factory.setNamespaceAware(true);

        final DocumentBuilder docBuilder = createDocBuilder(factory);
        return docBuilder.parse(new InputSource(in));
    }

    /**
     * Creates a document builder with the warning-aware error handler.
     *
     * @param factory document builder factory that creates the document builder
     * @return the created document builder
     * @throws ParserConfigurationException if document builder cannot be constructed with the specified parameters
     */
    private static DocumentBuilder createDocBuilder(DocumentBuilderFactory factory) throws ParserConfigurationException {
        DocumentBuilder documentBuilder = factory.newDocumentBuilder();

        documentBuilder.setErrorHandler(new ErrorHandler() {
            public void warning(SAXParseException exception) throws SAXException {
                throw exception;
            }

            public void error(SAXParseException exception) throws SAXException {
                throw exception;
            }

            public void fatalError(SAXParseException exception) throws SAXException {
                throw exception;
            }
        });

        return documentBuilder;
    }
}

/**
 * Creates a test schema tree by adding children to the given node. Children consist of
 * simple and complex elements; complex elements are visited by the specified test-creating visitor.
 */
abstract class AbstractTestCreator {
    /**
     * Returns test parameters with the value of depth set according to the position of the complex element in the row
     * and the fill pattern.
     *
     * @param params the initial test parameters
     * @param i      position of the complex element in the row
     * @return the test parameters that are equal to the initial except the depth which is set according to the fill style
     */
    protected abstract TestParameters getParamsWithDepth(TestParameters params, int i);

    protected void fillChildren(SchemaNode node, TestParameters params, int nSimpleElements, int nComplexElements, TreeVisitor<Void, TestParameters, IOException> testCreator) throws IOException {
        SchemaNode newNode = node;
        if (nComplexElements > 0) {
            switch (params.getPositionPattern()) {
            case BEGINNING:
                newNode = createNodeDeeperLevel(params.getNestingNode(), node, TEST_CODE);
                newNode.accept(testCreator, getParamsWithDepth(params, 0));
                for (int i = 1; i < nComplexElements; ++i) {
                    newNode = createNodeSameLevel(params.getNestingNode(), newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }

                for (int i = 0; i < nSimpleElements; ++i) {
                    newNode = createNodeSameLevel(NodeType.ACTION, newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }
                break;
            case MIDDLE:
                int mid = nSimpleElements / 2;
                newNode = createNodeDeeperLevel(NodeType.ACTION, node, TEST_CODE);
                newNode.accept(testCreator, getParamsWithDepth(params, 0));
                for (int i = 0; i < mid; ++i) {
                    newNode = createNodeSameLevel(NodeType.ACTION, newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }
                for (int i = 0; i < nComplexElements; ++i) {
                    newNode = createNodeSameLevel(params.getNestingNode(), newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }

                for (int i = mid + 1; i < nSimpleElements; ++i) {
                    newNode = createNodeSameLevel(NodeType.ACTION, newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }
                break;
            case END:
                newNode = createNodeDeeperLevel(NodeType.ACTION, node, TEST_CODE);
                newNode.accept(testCreator, getParamsWithDepth(params, 0));
                for (int i = 1; i < nSimpleElements; ++i) {
                    newNode = createNodeSameLevel(NodeType.ACTION, newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }
                for (int i = 0; i < nComplexElements; ++i) {
                    newNode = createNodeSameLevel(params.getNestingNode(), newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, i));
                }
                break;
            case RANDOM:
                boolean firstSimple = Util.randomBool();
                newNode = createNodeDeeperLevel(firstSimple ? NodeType.ACTION : params.getNestingNode(), node, TEST_CODE);
                newNode.accept(testCreator, getParamsWithDepth(params, 0));

                int simpleCount = firstSimple ? nSimpleElements - 1 : nSimpleElements;
                int complexCount = !firstSimple ? nComplexElements - 1 : nComplexElements;
                while (simpleCount > 0 || complexCount > 0) {
                    boolean simple;
                    if (simpleCount <= 0) {
                        simple = false;
                    } else if (complexCount <= 0) {
                        simple = true;
                    } else {
                        simple = Util.randomBool();
                    }

                    int dummy = simple ? --simpleCount : --complexCount;
                    newNode = createNodeSameLevel(simple ? NodeType.ACTION : params.getNestingNode(), newNode, TEST_CODE);
                    newNode.accept(testCreator, getParamsWithDepth(params, nSimpleElements - simpleCount - 1));
                }
                break;
            }
        } else {
            newNode = createNodeDeeperLevel(NodeType.ACTION, node, TEST_CODE);
            newNode.accept(testCreator, getParamsWithDepth(params, 0));
            for (int i = 1; i < nSimpleElements; ++i) {
                newNode = createNodeSameLevel(NodeType.ACTION, newNode, TEST_CODE);
                newNode.accept(testCreator, getParamsWithDepth(params, i));
            }
        }
        // add break if necessary
        boolean addBreak = false;
        switch (params.getBreakPattern()) {
        case EACH:
            addBreak = true;
            break;
        case NO:
            addBreak = false;
            break;
        case RANDOM:
            addBreak = Util.randomBool();
            break;
        }
        if (addBreak) {
            if (newNode != node) {
                newNode = createNodeSameLevel(params.getBreakNode(), newNode, TEST_CODE);
            } else {
                newNode = createNodeDeeperLevel(params.getBreakNode(), newNode, TEST_CODE);
            }
            newNode.accept(testCreator, getParamsWithDepth(params, nSimpleElements + nComplexElements - 1));
        }
    }
}

class FirstLevelTestCreator extends AbstractTestCreator {
    private final TestCreator testCreator = new TestCreator();

    public SchemaTree[] createTests(Element xmlNode) throws IOException {
        switch (NodeType.getNodeType(xmlNode.getNodeName())) {
        case BLOCK:
        case DO_WHILE:
        case FOR:
        case SYNCHRONIZED:
        case WHILE:
            return createNestedNodeTest(xmlNode);
        case IF:
        case SWITCH:
            return createBranchNodeTest(xmlNode);
        case TRY:
            return createTryNodeTest(xmlNode);
        }
        return new SchemaTree[0];
    }

    private SchemaTree[] createNestedNodeTest(Element xmlNode) throws IOException {
        Element nested = getNestedLevelsDescription(xmlNode);
        try {
            List<SchemaTree> retList = new ArrayList<SchemaTree>();
            for (TestParameters p : new TestParameters(Integer.valueOf(xmlNode.getAttribute("maxSimpleElements")), xmlNode, nested)) {
                SchemaNode root = SchemaNode.createNode(TRANSLATION_UNIT, null, null, TEST_CODE);
                SchemaNode node = newTest(xmlNode, root);
                fillChildren(node, p, p.getSimpleElements(), p.getComplexElements(), testCreator);
                retList.add(createAnnotatedSchemaTree(p, root));
            }
            return retList.toArray(new SchemaTree[retList.size()]);
        } catch (NumberFormatException e) {
            throw new IOException("error occurred while reading test properties: " + e.getMessage());
        }
    }

    /**
     * Helper method to create a new test.
     *
     * @param xmlNode <tt>XML</tt> node containing description of the test
     * @param root    the root schema node of the test
     * @return the first non-module node in the test
     */
    private SchemaNode newTest(Element xmlNode, SchemaNode root) {
        SchemaNode classNode = createNodeDeeperLevel(CLASS, root, TEST_CODE);
        SchemaNode methodNode = createNodeDeeperLevel(METHOD, classNode, TEST_CODE);
        return createNodeDeeperLevel(NodeType.getNodeType(xmlNode.getNodeName()), methodNode, TEST_CODE);
    }

    /**
     * Gets nested levels description  if present, it's the first child element.
     */
    private Element getNestedLevelsDescription(Element xmlNode) {
        Element nested = null;
        NodeList children = xmlNode.getChildNodes();
        for (int i = 0; i < children.getLength(); ++i) {
            if (children.item(i).getNodeType() == Node.ELEMENT_NODE) {
                nested = (Element) children.item(i);
                break;
            }
        }
        return nested;
    }

    private SchemaTree[] createBranchNodeTest(Element xmlNode) throws IOException {
        Element nested = getNestedLevelsDescription(xmlNode);
        FillPattern fillPattern = FillPattern.getValue(xmlNode.getAttribute("fill"));
        int nBranches = Integer.valueOf(xmlNode.getAttribute("branches"));
        int curBranches = 0;
        boolean iterate = Boolean.valueOf(xmlNode.getAttribute("iterate"));
        List<SchemaTree> retList = new ArrayList<SchemaTree>();
        for (TestParameters p : new TestParameters(Integer.valueOf(xmlNode.getAttribute("maxSimpleElements")), xmlNode, nested)) {
            SchemaNode root = SwitchNode.createNode(TRANSLATION_UNIT, null, null, TEST_CODE);
            SchemaNode node = newTest(xmlNode, root);
            if (curBranches < nBranches) {
                ++curBranches;
            }
            fillBranchNode((BranchNode) node, xmlNode, nested, fillPattern, p.getSimpleElements(), iterate ? curBranches : nBranches);
            retList.add(createAnnotatedSchemaTree(p, root));
        }
        return retList.toArray(new SchemaTree[retList.size()]);
    }

    private void fillBranchNode(BranchNode node, Element xmlNode, Element nested, FillPattern fillPattern, int maxSimpleElements, int nBranches) throws IOException {
        for (int i = 0; i < nBranches; ++i) {
            node.addNewConditionSet(TEST_CODE);
            try {
                TestParameters params = new TestParameters(fillPattern.getFilledValue(i, nBranches, maxSimpleElements), xmlNode, nested);
                fillChildren(node, params, params.getSimpleElements(), params.getComplexElements(), testCreator);
            } catch (IllegalArgumentException e) {
                throw new IOException("error occurred while reading test properties: " + e.getMessage());
            }
        }
    }

    private SchemaTree[] createTryNodeTest(Element xmlNode) throws IOException {
        NodeList tryNodes = xmlNode.getChildNodes();
        // we'll get schema trees for try, catch, and finally nodes (if possible)
        // and create tests from their combinations up to the specified amount
        SchemaTree[] tryNodeTrees = new SchemaTree[0];
        SchemaTree[] catchNodeTrees = new SchemaTree[0];
        SchemaTree[] finallyNodeTrees = new SchemaTree[0];

        // check that the first element node must be try-node
        Element tryNode = null;
        int i = 0;
        for (; i < tryNodes.getLength(); ++i) {
            Node cur = tryNodes.item(i);
            if (cur.getNodeType() == Node.ELEMENT_NODE && cur.getNodeName().equals("try")) {
                tryNode = (Element) tryNodes.item(i);
                tryNodeTrees = createNestedNodeTest(tryNode);
                break;
            }
        }
        if (tryNode == null) {
            throw new IOException("Illegal try-catch-finally construction.");
        }

        // visit the rest of the nodes (catch and finally, if present)
        SchemaNode lastNode = null;
        for (; i < tryNodes.getLength(); ++i) {
            Node cur = tryNodes.item(i);
            if (cur.getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            if (cur.getNodeName().equals("catch")) {
                catchNodeTrees = createBranchNodeTest((Element) cur);
            } else if (cur.getNodeName().equals("finally")) {
                finallyNodeTrees = createNestedNodeTest((Element) cur);
            }
        }

        // create combinations
        Integer nComb = Integer.valueOf(xmlNode.getAttribute("combinations"));
        SchemaTree[] res = new SchemaTree[nComb];
        for (i = 0; i < nComb; ++i) {
            int ind = Util.random(tryNodeTrees.length - 1);
            res[i] = tryNodeTrees[ind];
            // find the try node
            SchemaNode last = getNode(res[i], NodeType.TRY);
            if (catchNodeTrees.length > 0) {
                SchemaNode catchNode = getNode(catchNodeTrees[Util.random(catchNodeTrees.length - 1)], NodeType.CATCH);
                last.setNext(catchNode);
                last = catchNode;
            }
            if (finallyNodeTrees.length > 0) {
                SchemaNode finallyNode = getNode(finallyNodeTrees[Util.random(finallyNodeTrees.length - 1)], NodeType.FINALLY);
                last.setNext(finallyNode);
            }
        }
        return res;
    }


    /**
     * Gets the first node (counting from the root) in the specified schema tree.
     *
     * @param tree
     * @param type
     * @return
     */
    private SchemaNode getNode(SchemaTree tree, NodeType type) {
        SchemaNode node = tree.getRoot();
        while (node.getNodeType() != type) {
            if (node.hasChildren()) {
                node = node.getChildren().next();
            } else if (node.hasNext()) {
                node = node.getNext();
            } else {
                break;
            }
        }
        return node;
    }

//    public SchemaTree[] visit(TryNode node, Element xmlNode) throws IOException {
//        NodeList tryNodes = xmlNode.getChildNodes();
//        // the first element node must be try-node
//        Element tryNode = null;
//        int i = 0;
//        for (; i < tryNodes.getLength(); ++i) {
//            Node cur = tryNodes.item(i);
//            if (cur.getNodeType() == Node.ELEMENT_NODE && cur.getNodeName().equals("try-node")) {
//                    tryNode = (Element) tryNodes.item(i);
//                break;
//            }
//        }
//        if (tryNode == null) {
//            throw new IOException("Illegal try-catch-finally construction.");
//        }
//        visit((NestingNode) node, tryNode);
//
//        // visit the rest of the nodes (catch and finally, if present)
//        SchemaNode lastNode = node;
//        for (; i < tryNodes.getLength(); ++i) {
//            Node cur = tryNodes.item(i);
//            if (cur.getNodeType() != Node.ELEMENT_NODE) {
//                continue;
//            }
//            if (cur.getNodeName().equals("catch-node")) {
//                lastNode = createNodeSameLevel(NodeType.CATCH, lastNode, TEST_CODE);
//                visit((BranchNode) lastNode, (Element) cur);
//            } else if (cur.getNodeName().equals("finally-node")) {
//                lastNode = createNodeSameLevel(NodeType.FINALLY, lastNode, TEST_CODE);
//                visit((NestingNode) lastNode, (Element) cur);
//            }
//        }
//        return new SchemaTree[0];
//    }

    /**
     * Creates a schema tree with the description regarding the test properties used to generate the tree.
     *
     * @param p        test parameters
     * @param treeRoot root node of the tree
     * @return the created schema tree
     */
    private SchemaTree createAnnotatedSchemaTree(TestParameters p, SchemaNode treeRoot) {
        String descr = p.toString();
        CharArrayWriter charWriter = new CharArrayWriter();
        PrintWriter stringWriter = new PrintWriter(charWriter);
        TextExporter textExporter = new TextExporter();
        textExporter.export(stringWriter, new DepthFirstTreeWalker<SchemaNode>(treeRoot));
        descr += String.format("%n") + new String(charWriter.toCharArray());
        return new SchemaTree(treeRoot, descr);
    }

    protected TestParameters getParamsWithDepth(TestParameters params, int i) {
        int newDepth = params.getMaxDepth() == 0 ? 0 : params.getNestedPattern().getFilledValue(i, params.getComplexElements(), params.getMaxDepth());
        return new TestParameters(params, newDepth);
    }
}

class TestCreator extends AbstractTestCreator implements TreeVisitor<Void, TestParameters, IOException> {
    /**
     * Amount of branches for inner branch nodes.
     */
    private static final int BRANCHES = 2;

    public Void visit(SchemaNode node, TestParameters context) {
        return null;
    }

    public Void visit(BranchNode node, TestParameters context) throws IOException {
        if (context.getMaxDepth() == 0) {
            return null;
        }

        for (int i = 0; i < BRANCHES; ++i) {
            node.addNewConditionSet(TEST_CODE);
            fillChildren(node, context, context.getSimpleElementsInNested(), context.getComplexElementsInNested(), this);
        }
        return null;
    }

    public Void visit(NestingNode node, TestParameters context) throws IOException {
        if (context.getMaxDepth() == 0) {
            return null;
        }
        fillChildren(node, context, context.getSimpleElementsInNested(), context.getComplexElementsInNested(), this);
        return null;
    }

    public Void visit(FlowChangeNode node, TestParameters context) {
        return null;
    }

    public Void visit(TryNode node, TestParameters context) {
        if (context.getMaxDepth() == 0) {
            return null;
        }
        return null;
    }

    protected TestParameters getParamsWithDepth(TestParameters params, int i) {
        return new TestParameters(params, params.getNestedPattern().getFilledValue(i, params.getComplexElementsInNested(), params.getMaxDepth() - 1));
    }
}

