package ru.amse.baltijsky.javascheme.importer.test;

import org.w3c.dom.Element;
import ru.amse.baltijsky.javascheme.tree.NodeType;
import ru.amse.baltijsky.javascheme.util.Util;

import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;

/**
 * Parameters for a node test  describes the structure of the test.
 * Iterator provides opportunity to create iterated tests with varying test parameters.
 */
class TestParameters implements Iterable<TestParameters> {
    /**
     * Name of the operator for which the test is created.
     */
    private final String name;
    /**
     * Amount of simple elements in the first nested level or the characteristic amount of simple elements in a branch.
     */
    private int simpleElements;
    /**
     * Amound of complex elements in the first nested level or the characteristic amount of simple elements in a branch.
     */
    private int complexElements = 0;
    /**
     * Rule for distributing break nodes over branches.
     */
    private BreakPositionPattern breakPattern;
    /**
     * The type of the break node. Can be null, in this case random node is returned.
     */
    private NodeType breakNode = null;
    /**
     * Determines whether the test has nested levels.
     */
    private final boolean hasNested;
    /**
     * Pattern for positioning complex elements amoung simple elements.
     */
    private PositionPattern positionPattern;
    /**
     * Maximum amount of nested levels in complex constructions.
     */
    private int maxDepth = 0;
    /**
     * Pattern of depth distrubution among complex constructions.
     */
    private FillPattern nestedPattern;
    /**
     * Type of the nodes with nested levels. Can be null, in this case random node is returned.
     */
    private NodeType nestingNode = null;
    /**
     * Amount of simple elements in the nested nodes.
     */
    private int simpleElementsInNested = 1;
    /**
     * Amount of complex elements (<i>= have nested levels<i>) in the nested nodes.
     */
    private int complexElementsInNested = 1;
    /**
     * If true, iteration over TestParameters returns test parameters with increasing properties values.
     */
    private boolean iterate;

    /**
     * Creates test parameters from the description in the xml DOM element.
     *
     * @param simpleElements amount of simple elements in the first level of node children
     * @param element        DOM element which attributes describe simple elements in the first level of node children
     * @param nested         DOM element which attributes describe behavior in all levels of node children (starting with
     *                       the amount of complex elements in the first level of node children)
     */
    public TestParameters(int simpleElements, Element element, Element nested) {
        this.simpleElements = simpleElements;
        name = element.getNodeName();
        complexElements = 0;
        // set general information
        breakPattern = BreakPositionPattern.getValue(element.getAttribute("breakPattern"));
        iterate = Boolean.valueOf(element.getAttribute("iterate"));
        try {
            breakNode = NodeType.getNodeType(element.getAttribute("breakNode"));
        } catch (IllegalArgumentException e) {
            breakNode = null;
        }

        // set information about complex elements
        if (!(hasNested = (nested != null))) {
            return;
        }
        complexElements = Integer.valueOf(nested.getAttribute("nestedElements"));
        maxDepth = Integer.valueOf(nested.getAttribute("maxNestedDepth"));
        nestedPattern = FillPattern.getValue((nested.getAttribute("nestedPattern")));
        positionPattern = PositionPattern.getValue(nested.getAttribute("positionPattern"));
        try {
            nestingNode = NodeType.getNodeType(nested.getAttribute("nestedType"));
        } catch (IllegalArgumentException e) {
            nestingNode = null;
        }
        simpleElementsInNested = Integer.valueOf(nested.getAttribute("simpleElementsInNested"));
        complexElementsInNested = Integer.valueOf(nested.getAttribute("complexElementsInNested"));
    }

    public TestParameters(TestParameters o, int maxDepth) {
        name = o.name;
        breakNode = o.breakNode;
        breakPattern = o.breakPattern;
        complexElements = o.complexElements;
        complexElementsInNested = o.complexElementsInNested;
        hasNested = o.hasNested;
        this.maxDepth = maxDepth;
        nestedPattern = o.nestedPattern;
        nestingNode = o.nestingNode;
        positionPattern = o.positionPattern;
        simpleElements = o.simpleElements;
        simpleElementsInNested = o.simpleElementsInNested;
    }

    public int getSimpleElements() {
        return simpleElements;
    }

    public int getComplexElements() {
        return complexElements;
    }

    public BreakPositionPattern getBreakPattern() {
        return breakPattern;
    }

    public NodeType getBreakNode() {
        if (breakNode != null) {
            return breakNode;
        }
        int rand = Util.random(4);
        switch (rand) {
        case 0:
            return NodeType.BREAK;
        case 1:
            return NodeType.CONTINUE;
        case 2:
            return NodeType.RETURN;
        default:
            return NodeType.THROW;
        }
    }

    public PositionPattern getPositionPattern() {
        return positionPattern;
    }

    public int getMaxDepth() {
        return maxDepth;
    }

    public FillPattern getNestedPattern() {
        return nestedPattern;
    }

    public NodeType getNestingNode() {
        if (nestingNode != null) {
            return nestingNode;
        }
        int rand = Util.random(8);
        switch (rand) {
        case 0:
            return NodeType.BLOCK;
        case 1:
            return NodeType.IF;
        case 2:
            return NodeType.SWITCH;
        case 3:
            return NodeType.FOR;
        case 4:
            return NodeType.WHILE;
        case 5:
            return NodeType.DO_WHILE;
        case 6:
            return NodeType.SYNCHRONIZED;
        default:
            return NodeType.TRY;
        }
    }

    public int getSimpleElementsInNested() {
        return simpleElementsInNested;
    }

    public int getComplexElementsInNested() {
        return complexElementsInNested;
    }

    public Iterator<TestParameters> iterator() {
        if (iterate) {
            return new IteratedTestParameters(this);
        } else {
            return Collections.singleton(this).iterator();
        }
    }

    /**
     * Returns textual desription of the test defined by these test parameters.
     *
     * @return description of the test corresponding to test parameters
     */
    public String toString() {
        StringBuilder ret = new StringBuilder();
        final String nl = String.format("%n");
        ret.append("main operator: ").append(name).append(nl);
        ret.append("iterate: ").append(iterate).append(nl);
        ret.append("simple elements on the outer level: ").append(simpleElements).append(nl);
        ret.append("complex elements on the outer level: ").append(complexElements).append(nl);
        ret.append("break pattern: ").append(breakPattern.name()).append(nl);
        ret.append("break node: ").append(breakNode == null ? "random" : breakNode.name()).append(nl);
        ret.append("nested types included: ").append(hasNested).append(nl);
        if (hasNested) {
            ret.append("position pattern: ").append(positionPattern.name()).append(nl);
            ret.append("maximum depth: ").append(maxDepth).append(nl);
            ret.append("nested pattern: ").append(nestedPattern.name()).append(nl);
            ret.append("simple elements in nested levels: ").append(simpleElementsInNested).append(nl);
            ret.append("complex elements in nested levels: ").append(complexElementsInNested).append(nl);
        }
        return ret.toString();
    }

    /**
     * For each invokation of <code>next()</code> returns test parameters with increased properties.
     */
    private class IteratedTestParameters implements Iterator<TestParameters> {

        /**
         * Current value of properties (increased with each call to <code>next()</code>). It is distributed over the
         * test parameters in a saturated way: if a value reaches the value in the host test parameters, it remains
         * the same until the end of iteration.
         */
        private int curValue = 0;

        /**
         * Maximum value  marks the end of the iteration.
         */
        private int maxValue;

        private TestParameters lastTestParams;

        private TestParameters hostTestParams;

        private IteratedTestParameters(TestParameters host) {
            maxValue = simpleElements;
            if (maxValue < complexElements) maxValue = complexElements;
            if (maxValue < simpleElementsInNested) maxValue = simpleElementsInNested;
            if (maxValue < complexElementsInNested) maxValue = complexElementsInNested;
            if (maxValue < maxDepth) maxValue = maxDepth;

            hostTestParams = host;
            lastTestParams = new TestParameters(host, 0);
            lastTestParams.simpleElements = 0;
            lastTestParams.complexElements = 0;
            lastTestParams.simpleElementsInNested = 0;
            lastTestParams.complexElementsInNested = 0;
        }

        public boolean hasNext() {
            return curValue < maxValue;
        }

        public TestParameters next() {
            if (!hasNext()) {
                throw new IllegalStateException();
            }
            ++curValue;
            lastTestParams.simpleElements = curValue <= hostTestParams.simpleElements ? curValue : hostTestParams.simpleElements;
            lastTestParams.complexElements = curValue <= hostTestParams.complexElements ? curValue : hostTestParams.complexElements;
            lastTestParams.simpleElementsInNested = curValue <= hostTestParams.simpleElementsInNested ? curValue : hostTestParams.simpleElementsInNested;
            lastTestParams.complexElementsInNested = curValue <= hostTestParams.complexElementsInNested ? curValue : hostTestParams.complexElementsInNested;
            lastTestParams.maxDepth = curValue <= hostTestParams.maxDepth ? curValue : hostTestParams.maxDepth;
            return lastTestParams;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }

    }
}

/**
 * Fill pattern for elements.
 */
enum FillPattern {
    EQUAL, ASCENDING, DESCENDING, RANDOM;

    public static FillPattern getValue(String s) {
        return Util.getEnumValue(s, Arrays.asList(values()));
    }

    /**
     * Returns the value set according to the current fill pattern and its position.
     *
     * @param i        row position of the element for which the value is calculated
     * @param amount   amount of elements in the row
     * @param maxValue maximum allowed value
     * @return the value for the specified position according to the fill pattern
     */
    public int getFilledValue(int i, int amount, int maxValue) {
        if (maxValue <= 0) {
            return 0;
        }

        int value = 0;
        switch (this) {
        case EQUAL:
            value = maxValue;
            break;
        case ASCENDING:
            value = (i + 1 <= maxValue) ? i + 1 : maxValue;
            break;
        case DESCENDING:
            value = amount - i;
            if (value < 1) {
                value = 1;
            }
            break;
        case RANDOM:
            value = Util.random(maxValue);
            break;
        }
        return value;
    }
}

/**
 * Position pattern for complex elements among simple ones.
 */
enum PositionPattern {
    BEGINNING, MIDDLE, END, RANDOM;

    public static PositionPattern getValue(String s) {
        return Util.getEnumValue(s, Arrays.asList(values()));
    }
}

/**
 * Break position pattern.
 */
enum BreakPositionPattern {
    NO, EACH, RANDOM;

    public static BreakPositionPattern getValue(String s) {
        return Util.getEnumValue(s, Arrays.asList(values()));
    }
}
